from enum import IntEnum


class StatusTypes(IntEnum):
    ACTIVE = 1
    PENDING = 2
    DELETED = 3
    COMPLETED = 4
    FAILED = 5
    PROCESSING = 6
    DISAPPROVED = 7
    PAYMENTPENDING = 8
    
    @classmethod
    def choices(cls):
        return [(key.value, key.name) for key in cls]


class NetworkTypes(IntEnum):
    MTN = 0
    VODAFONE = 1
    AIRTELTIGO = 2
    GLO = 3
    EXPRESSO = 4
    UNKNOWN = 5

    @classmethod
    def choices(cls):
        return [(key.value, key.name) for key in cls]

    @classmethod
    def choicesAsDict(cls):
        return {key.value: key.name for key in cls}
    
    @classmethod
    def dichoices(cls):
        return {key.name:key.value for key in cls}


from django.db import models
from django.contrib.auth.models import User
from . import enum


class MomoTransaction(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    amount = models.DecimalField(max_digits=10, decimal_places=4)
    network = models.IntegerField(choices=enum.NetworkTypes.choices())
    msisdn = models.CharField(max_length=20)
    merchant_id = models.CharField(max_length=20)
    merchant_name = models.CharField(max_length=100)
    status = models.IntegerField(choices=enum.StatusTypes.choices(), default=enum.StatusTypes.PENDING)
    info_sent = models.TextField()
    initial_callback = models.CharField(max_length=255, null=True)
    final_callback = models.CharField(max_length=255, null=True)
    time_created = models.DateTimeField(auto_now_add=True)
    time_updated = models.DateTimeField(auto_now=True)
    order_id = models.CharField(max_length=50, unique=True)
    ext_id = models.CharField(max_length=50,null=True)


from django.conf import settings
from .models import MomoTransaction
import string, random, json, requests, hashlib
from .enum import NetworkTypes, StatusTypes
from django.core.serializers.json import DjangoJSONEncoder 

class MomoPayProcessor:
    
    def __init__(self, data=None, merchant=None, callback=None, *args, **kwargs):

        if data is not None:
            self.user = data['user']
            self.amount = data['amount']
            self.momo_number = data['customerNumber']
            self.network = data['payby']
            self.item_desc = data['item_desc']
            self.orderID = data.get('order_id', self.generateOrderID())
            self.network_name = NetworkTypes(data['payby']).name
        
        if callback is not None:
            self.callback = callback
        
        self.merchant = merchant

    def pay(self):
        set_merchant = self.setMerchant()
        if set_merchant['status']:
            self.getPaymentPayload()
            try:
                self.momotrans = MomoTransaction.objects.create(
                    user = self.user,
                    amount = self.amount,
                    network = self.network,
                    msisdn = self.momo_number,
                    info_sent = self.payload,
                    order_id = self.orderID,
                    merchant_id = self.merchantID,
                    merchant_name = self.merchant_username,
                )
            except Exception as e:
                return {"status":False, "message":f"{e}"}
            self.response = requests.post(settings.PAYMENT_GATEWAY, data=json.dumps(self.payload, cls=DjangoJSONEncoder))
            self.momotrans.initial_callback = self.response.text
            self.momotrans.save()
            self.parseMomoResponse()
        return set_merchant

        

    def parseMomoResponse(self):

        try:
            response = self.response.json()
            
            if response["Status"] == "Accepted":
                self.momotrans.ext_id = response["InvoiceNo"]
                self.momotrans.save()
                return {"status":True, "message":"Passed"}

            self.momotrans.final_callback = response.text
            
        except Exception as e:
            self.momotrans.final_callback = f"{e}"
        
        self.momotrans.status = StatusTypes.FAILED
        self.momotrans.save()
        return {"status": False, "message":"Failed"}

    
    def setMerchant(self):

        if self.merchant is not None:
            self.merchantID = merchant.get('id')
            self.merchant_username = merchant.get('username')
            self.merchant_password = merchant.get('password')
        else:
            try:
                self.merchantID = settings.PAYMENT_MERCHANT
                self.merchant_username = settings.PAYMENT_USERNAME
                self.merchant_password = settings.PAYMENT_PASSWORD
            except Exception as e:
                return {"status":False, "message":f"{e}"}
        return {"status":True}


    
    def generateOrderID(self):
        alphabet = string.ascii_uppercase + string.digits
        return ''.join(random.choices(alphabet, k=12))


    def getPaymentPayload(self):
        key = random.randint(1000, 9999)
        secret = hashlib.md5((self.merchant_username+str(key)+self.merchant_password).encode("utf-8")).hexdigest()
        self.payload = {
            'merchant_id': self.merchantID,
            'secrete': secret,
            'key': key,
            'item_desc': self.item_desc,
            'callback': settings.PAYMENT_CALLBACK,
            'order_id': self.orderID,
            'customerName': self.user.username,
            'customerNumber': self.momo_number,
            'payby': self.network_name,
            'amount': self.amount,
        }
        if self.network == NetworkTypes.VODAFONE:
            self.payload["newVodaPayment"] = True



    def onCallback(self):
        if self.callback is not None:
        
            if self.callback["Status"] == "PAID":
                self.status_type = StatusTypes.COMPLETED
            else:
                self.status_type = StatusTypes.FAILED

            try:
                momo_trans = MomoTransaction.objects.get(
                    order_id=self.callback["Order_id"],
                    ext_id=self.callback["InvoiceNo"])
                momo_trans.final_callback = json.dumps(self.callback , cls=DjangoJSONEncoder)
                momo_trans.status = self.status_type                
                momo_trans.save()
                return {"status":True, "transactionID":momo_trans.id, "amount":momo_trans.amount}
            except Exception as e:
                return {"status":False, "message":f"{e}"}


from django.urls import path, include
from . import views

urlpatterns = [
    path('hello/', views.helloView, name='hello'),
    path('user/<int:id>/',views.MomoPaymentView.as_view()),
    path('callback/',views.MomoPaymentCallbackView.as_view()),
]


from django.contrib import admin
from django.apps import apps
models = apps.get_models()

# Register your models here.

for model in models:
    try:
        admin.site.register(model)
    except admin.sites.AlreadyRegistered:
        pass


# Generated by Django 2.2.10 on 2020-08-12 16:50

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
from .. import enum


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [    
        migrations.CreateModel(
            name='MomoTransaction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('amount', models.DecimalField(decimal_places=4, max_digits=10)),
                ('network', models.IntegerField(choices=[(0, 'MTN'), (1, 'VODAFONE'), (2, 'AIRTELTIGO'), (3, 'GLO'), (4, 'EXPRESSO'), (5, 'UNKNOWN')])),
                ('msisdn', models.CharField(max_length=20)),
                ('merchant_id', models.CharField(max_length=20)),
                ('merchant_name', models.CharField(max_length=100)),
                ('status', models.IntegerField(choices=[(1, 'ACTIVE'), (2, 'PENDING'), (3, 'DELETED'), (4, 'COMPLETED'), (5, 'FAILED'), (6, 'PROCESSING'), (7, 'DISAPPROVED'), (8, 'PAYMENTPENDING')], default=enum.StatusTypes(2))),
                ('info_sent', models.TextField()),
                ('initial_callback', models.CharField(max_length=255, null=True)),
                ('final_callback', models.CharField(max_length=255, null=True)),
                ('time_created', models.DateTimeField(auto_now_add=True)),
                ('time_updated', models.DateTimeField(auto_now=True)),
                ('order_id', models.CharField(max_length=50, unique=True)),
                ('ext_id', models.CharField(max_length=50, null=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



from django.shortcuts import render
from django.http import HttpResponse
from rest_framework.views import APIView
from rest_framework.status import HTTP_400_BAD_REQUEST, HTTP_200_OK
from rest_framework.response import Response
from .serializer import MomoPaymentSerializer, MomoCallbackSerializer
from .momopay import MomoPayProcessor

# Create your views here.
def helloView(request):
    return HttpResponse('Hello, World!')


class MomoPaymentView(APIView):
    def post(self, request, id, *args, **kwargs):
        request_data = {'user': id}
        request_data.update(request.data)
        momoPaymentSerializer = MomoPaymentSerializer(data=request_data)
        if momoPaymentSerializer.is_valid():
            response = MomoPayProcessor(data=momoPaymentSerializer.validated_data).pay()
            return Response(response, status=HTTP_200_OK)
        return Response(momoPaymentSerializer.errors, status=HTTP_400_BAD_REQUEST)


class MomoPaymentCallbackView(APIView):
    def post(self, request):
        serializerResult = MomoCallbackSerializer(data=request.data)
        if serializerResult.is_valid(raise_exception=True):
            response = MomoPayProcessor(callback=serializerResult.validated_data).onCallback()
            return Response(response, status=HTTP_200_OK)
        return Response(serializerResult.errors, status=HTTP_400_BAD_REQUEST)


from rest_framework import serializers
from .enum import NetworkTypes
from django.contrib.auth.models import User


class MomoPaymentSerializer(serializers.Serializer):

    user = serializers.IntegerField(required=True)
    amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    item_desc = serializers.CharField(required=True)
    customerNumber = serializers.CharField(required=True)
    payby = serializers.CharField(required=True)
    

    def validate_item_desc(self, value):
        item_desc_len = len(value)
        if item_desc_len < 10  or item_desc_len > 24 :
            raise serializers.ValidationError("Item description should be greater than 9 but less than 25 characters")
        return value
    
    def validate_newVodaPayment(self, value):
        if type(value) is not bool:
            raise serializers.ValidationError("Boolean is required")
        return value

    
    def validate_customerNumber(self, value):
        if not value.isdigit() or len(value) < 10:
            raise serializers.ValidationError("Number should be digits only")
        return value

    
    def validate_payby(self, network_name):
        try:
            payby = NetworkTypes.dichoices()[network_name.upper()]
            return payby
        except KeyError:
            raise serializers.ValidationError("Sorry, invalid network")


    def validate_user(self, userID):
        try:
            user = User.objects.get(pk=userID)
        except:
            raise serializers.ValidationError("Sorry, user does not exist")
        return user

class MomoCallbackSerializer(serializers.Serializer):
    Timestamp = serializers.DateTimeField(format="%Y-%m-%d %H:%M:%S")
    Status = serializers.CharField()
    InvoiceNo = serializers.CharField()
    Order_id = serializers.CharField()